import string
import random
import logging

from django.contrib.sites.models import Site

from brevisurl.backends.base import BaseBrevisUrlBackend
from brevisurl.models import ShortUrl


log = logging.getLogger(__name__)


class BrevisUrlBackend(BaseBrevisUrlBackend):

    PROTOCOL = 'http'

    def shorten_url(self, original_url):
        """
        :raises: ImproperlyConfigured, django.core.exceptions.ValidationError
        """
        try:
            short_url = ShortUrl.objects.get(backend=self.class_path, original_url=original_url)
            log.info('Url "%s" already shortened to "%s"', original_url, short_url.shortened_url)
            return short_url
        except ShortUrl.DoesNotExist:
            pass

        try:
            current_site = Site.objects.get_current()
            short_url = ShortUrl()
            short_url.original_url = original_url
            short_url.shortened_url = '{0}://{1}/{2}'.format(self.PROTOCOL, current_site.domain,
                                                             self.__generate_token())
            short_url.backend = self.class_path
            short_url.save()
            log.info('Url "%s" shortened to "%s"', original_url, short_url.shortened_url)
            return short_url
        except Exception:
            if self.fail_silently:
                return None
            else:
                log.exception('Unknown exception raised while shortening url "%s"', original_url)
                raise

    def __generate_token(self, size=5):
        chars = list(string.ascii_letters + string.digits)
        random.shuffle(chars)
        while True:
            token = ''.join([random.choice(chars) for x in range(size)])
            if not ShortUrl.objects.filter(backend=self.class_path, shortened_url__endswith=token).count():
                break
        return token


from brevisurl.tests.test_brevisurl import *
from brevisurl.tests.test_utils import *
from brevisurl.tests.test_views import *
from brevisurl.tests.test_models import *
from brevisurl.tests.backends.test_local import *
from brevisurl.tests.templatetags.test_brevisurltags import *


import urlparse

from django import template
from django.template.defaulttags import URLNode, url
from django.contrib.sites.models import Site

from brevisurl import shorten_url as shorten_url_util, get_connection


register = template.Library()


@register.filter
def shorten_url(original_url):
    short_url = shorten_url_util(original_url, connection=get_connection(fail_silently=True))
    return short_url.shortened_url if short_url is not None else original_url


# Modified copy of django snippet #1518 (http://djangosnippets.org/snippets/1518/)
class AbsoluteURLNode(URLNode):
    def render(self, context):
        path = super(AbsoluteURLNode, self).render(context)
        domain = "http://{0}".format(Site.objects.get_current().domain)
        if self.asvar:
            context[self.asvar]= urlparse.urljoin(domain, context[self.asvar])
            return ''
        else:
            return urlparse.urljoin(domain, path)

def absurl(parser, token, node_cls=AbsoluteURLNode):
    """Just like {% url %} but ads the domain of the current site."""
    node_instance = url(parser, token)
    return node_cls(view_name=node_instance.view_name,
        args=node_instance.args,
        kwargs=node_instance.kwargs,
        asvar=node_instance.asvar)
absurl = register.tag(absurl)


from django.core.validators import URLValidator
from django.test import TestCase
from django.template import Template, Context

from brevisurl.models import ShortUrl


class TestShortenUrlTag(TestCase):

    def test_shorten_url_filter(self):
        original_url = 'http://www.codescale.net/'
        url = Template("""
        {% load brevisurltags %}
        {{ url|shorten_url }}
        """).render(Context({'url': original_url})).strip()
        self.assertEqual(ShortUrl.objects.all().count(), 1)
        self.assertEqual(ShortUrl.objects.all()[0].original_url, original_url)
        self.assertRegexpMatches(url, URLValidator.regex)

    def test_shorten_url_filter_invalid_url(self):
        original_url = 'www.codescale.'
        url = Template("""
        {% load brevisurltags %}
        {{ url|shorten_url }}
        """).render(Context({'url': original_url})).strip()
        self.assertEqual(ShortUrl.objects.all().count(), 0)
        self.assertEqual(url, original_url)

    def test_absurl_tag(self):
        url = Template("""
        {% load brevisurltags %}
        {% absurl brevisurl_redirect token='12345' as brevis_url %}
        {{ brevis_url|shorten_url }}
        """).render(Context()).strip()
        self.assertEqual(ShortUrl.objects.all().count(), 1)
        self.assertRegexpMatches(url, URLValidator.regex)


from django.test import TestCase
from django.test.client import Client
from django.core.urlresolvers import reverse
from django.contrib.sites.models import Site

from brevisurl import get_connection
from brevisurl.models import ShortUrl


class TestBrevisUrlRedirectView(TestCase):

    def setUp(self):
        self.site = Site.objects.get_current()
        self.connection = get_connection('brevisurl.backends.local.BrevisUrlBackend')
        self.short_url = ShortUrl()
        self.short_url.original_url = 'http://www.codescale.net/'
        self.short_url.shortened_url = '{0}://{1}/12345'.format(self.connection.PROTOCOL, self.site.domain)
        self.short_url.backend = self.connection.class_path
        self.short_url.save()
        self.client = Client()

    def test_redirect(self):
        response = self.client.get(reverse('brevisurl_redirect', kwargs={'token': 12345}))
        self.assertEqual(response.status_code, 301)

    def test_redirect_non_existing_token(self):
        response = self.client.get(reverse('brevisurl_redirect', kwargs={'token': 54321}))
        self.assertEqual(response.status_code, 404)


from django.utils import unittest

from brevisurl.utils import load_object
from brevisurl.backends.local import BrevisUrlBackend

class TestLoadObject(unittest.TestCase):

    def test_load_valid_object(self):
        klass = load_object('brevisurl.backends.local.BrevisUrlBackend')
        self.assertEqual(klass, BrevisUrlBackend)

    def test_load_import_error(self):
        with self.assertRaises(ImportError):
            load_object('brevisurl.local.BrevisUrlBackend')

    def test_load_attribute_error(self):
        with self.assertRaises(AttributeError):
            load_object('brevisurl.backends.local.NonExistingBackend')

    def test_load_value_error(self):
        with self.assertRaises(ValueError):
            load_object('brevisurl')


import brevisurl.settings
from brevisurl.utils import load_object


def get_connection(backend=None, fail_silently=False, **kwargs):
    """Load a brevisurl backend and return an instance of it.

    If backend is None (default) settings.BREVISURL_BACKEND is used.
    Both fail_silently and other keyword arguments are used in the
    constructor of the backend.

    :param backend: path to brevisurl backend
    :type backend: string
    :param fail_silently: whether to fail silently when error intercepted or not
    :type fail_silently: bool
    :returns: instance os BaseBrevisUrlBackend
    :rtype: brevisurl.backends.base.BaseBrevisUrlBackend

    """
    path = backend or brevisurl.settings.DEFAULT_BACKEND
    klass = load_object(path)
    return klass(fail_silently=fail_silently, **kwargs)


def shorten_url(original_url, fail_silently=False, connection=None):
    """Shortcut util for shortening urls using default brevisurl backend if none supplied.

    :param original_url: url that will be shortened
    :type original_url: string
    :param backend: one of brevisurl backends
    :type backend: brevisurl.backends.BaseBrevisUrlBackend
    :returns: shortened url from original url
    :rtype: brevisurl.models.ShortUrl

    """
    connection = connection or get_connection(fail_silently=fail_silently)
    return connection.shorten_url(original_url)


import logging

from django.http import Http404
from django.contrib.sites.models import Site
from django.views.generic.base import RedirectView
from django.core.exceptions import ImproperlyConfigured

from brevisurl import get_connection
from brevisurl.models import ShortUrl


log = logging.getLogger(__name__)


class BrevisUrlRedirectView(RedirectView):

    def get_redirect_url(self, **kwargs):
        try:
            token = kwargs.pop('token')
        except KeyError:
            log.exception('Token not found in keyword arguments')
            raise Http404
        connection = get_connection('brevisurl.backends.local.BrevisUrlBackend')
        protocol = connection.PROTOCOL
        try:
            site = Site.objects.get_current()
        except ImproperlyConfigured:
            log.exception('No site object configured for this django project')
            raise Http404
        try:
            short_url = '{0}://{1}/{2}'.format(protocol, site.domain, token)
            short_url_obj = ShortUrl.objects.get(backend=connection.class_path, shortened_url=short_url)
        except ShortUrl.DoesNotExist:
            log.exception('No shortened url found for backend: "%s" and token: "%s"', connection.class_path, token)
            raise Http404
        return short_url_obj.original_url


import hashlib
import logging
from django.core.exceptions import ValidationError

from django.db import models
from django.core.validators import URLValidator

from brevisurl import get_connection


log = logging.getLogger(__name__)


class ShortUrl(models.Model):
    """Model that represents shortened url."""
    original_url = models.URLField(null=False, blank=False)
    original_url_hash = models.CharField(max_length=64, null=False, blank=False)
    shortened_url = models.URLField(max_length=200, null=False, blank=False, unique=True)
    backend = models.CharField(max_length=200, null=False, blank=False)
    created = models.DateTimeField(auto_now_add=True, db_index=True, null=False, blank=False)

    def __unicode__(self):
        return self.shortened_url

    def get_connection(self, fail_silently=False):
        if not hasattr(self, 'brevis_connection'):
            if self.pk is not None:
                self.brevis_connection = get_connection(backend=self.backend, fail_silently=fail_silently)
            else:
                self.brevis_connection = get_connection(fail_silently=fail_silently)
        return self.brevis_connection

    def clean(self):
        url_validator = URLValidator()
        try:
            url_validator(self.original_url)
        except ValidationError:
            log.exception('ShortUrl.original_url is not valid URL')
            raise
        try:
            url_validator(self.shortened_url)
        except ValidationError:
            log.exception('ShortUrl.shortened_url is not valid URL')
            raise
        return super(ShortUrl, self).clean()

    def save(self, force_insert=False, force_update=False, using=None):
        if self.pk is None:
            self.original_url_hash = hashlib.sha256(self.original_url).hexdigest()
        self.full_clean()
        return super(ShortUrl, self).save(force_insert, force_update, using)

    class Meta:
        unique_together = (('original_url_hash', 'backend'),)
        verbose_name = 'Short url'
        verbose_name_plural = 'Short urls'
        ordering = ['-created']
        get_latest_by = 'created'


from abc import ABCMeta


class BaseBrevisUrlBackend(object):
    """Base class for brevisurl backend implementations. Subclasses must at least overwrite shorten_url()."""

    __metaclass__ = ABCMeta


    def __init__(self, fail_silently=False, **kwargs):
        self.fail_silently = fail_silently
        self.class_path = '{0}.{1}'.format(self.__module__, self.__class__.__name__)

    def open(self):
        """Open a network connection.

        This method can be overwritten by backend implementations to
        open a network connection.

        It's up to the backend implementation to track the status of
        a network connection if it's needed by the backend.

        The default implementation does nothing.

        """
        pass

    def close(self):
        """Close a network connection."""
        pass

    def shorten_url(self, original_url):
        """Shortens url into more compact form.

        :param original_url: url that will be shortened
        :type original_url: string
        :returns: shortened url from original url
        :rtype: brevisurl.models.ShortUrl

        """
        raise NotImplementedError


from django.utils import unittest

from brevisurl import get_connection, shorten_url
from brevisurl.backends.local import BrevisUrlBackend
from brevisurl.backends.base import BaseBrevisUrlBackend


class TestGetConnection(unittest.TestCase):

    def test_get_default_connection(self):
        connection = get_connection()
        self.assertIsInstance(connection, BrevisUrlBackend)

    def test_get_custom_connection(self):
        base_connection = get_connection(backend='brevisurl.backends.base.BaseBrevisUrlBackend')
        local_connection = get_connection(backend='brevisurl.backends.local.BrevisUrlBackend')
        self.assertIsInstance(base_connection, BaseBrevisUrlBackend)
        self.assertIsInstance(local_connection, BrevisUrlBackend)

    def test_get_connection_non_existing_backend(self):
        with self.assertRaises(AttributeError):
            get_connection(backend='brevisurl.backends.local.NonExistingBackend')


from django.conf.urls import patterns, url

from brevisurl import views


urlpatterns = patterns('brevisurl.views',
    url(r'^(?P<token>[a-zA-Z0-9]{5})$', views.BrevisUrlRedirectView.as_view(), name='brevisurl_redirect'),
)


from django.core.exceptions import ValidationError
from django.test import TestCase
from django.core.validators import URLValidator

from brevisurl import get_connection
from brevisurl.models import ShortUrl


class TestLocalBrevisUrlBackend(TestCase):

    def setUp(self):
        self.connection = get_connection('brevisurl.backends.local.BrevisUrlBackend')

    def test_shorten_url(self):
        self.assertEqual(ShortUrl.objects.all().count(), 0)
        original_url = 'http://www.codescale.net/'
        short_url = self.connection.shorten_url(original_url)
        self.assertEqual(ShortUrl.objects.all().count(), 1)
        self.assertEqual(short_url.original_url, original_url)
        self.assertRegexpMatches(short_url.shortened_url, URLValidator.regex)

    def test_shorten_url_reuse_old(self):
        original_url = 'http://www.codescale.net/'
        self.assertEqual(ShortUrl.objects.all().count(), 0)
        short_url = self.connection.shorten_url(original_url)
        self.assertEqual(ShortUrl.objects.all().count(), 1)
        self.assertEqual(short_url.original_url, original_url)
        self.assertRegexpMatches(short_url.shortened_url, URLValidator.regex)
        short_url = self.connection.shorten_url(original_url)
        self.assertEqual(ShortUrl.objects.all().count(), 1)
        self.assertEqual(short_url.original_url, original_url)
        self.assertRegexpMatches(short_url.shortened_url, URLValidator.regex)

    def test_shorten_url_create_new(self):
        original_url = 'http://www.codescale.net/'
        self.assertEqual(ShortUrl.objects.all().count(), 0)
        short_url = self.connection.shorten_url(original_url)
        self.assertEqual(ShortUrl.objects.all().count(), 1)
        self.assertEqual(short_url.original_url, original_url)
        self.assertRegexpMatches(short_url.shortened_url, URLValidator.regex)
        original_url = 'http://www.codescale.net/en/company/'
        short_url = self.connection.shorten_url(original_url)
        self.assertEqual(ShortUrl.objects.all().count(), 2)
        self.assertEqual(short_url.original_url, original_url)
        self.assertRegexpMatches(short_url.shortened_url, URLValidator.regex)

    def test_shorten_url_invalid_original_url(self):
        with self.assertRaises(ValidationError):
            self.connection.shorten_url('www.codescale.')
        self.assertEqual(ShortUrl.objects.all().count(), 0)

    def test_shorten_url_invalid_original_url_fail_silently(self):
        self.connection.fail_silently = True
        shorl_url = self.connection.shorten_url('www.codescale.')
        self.assertIsNone(shorl_url)


from django.utils import importlib


def load_object(import_path):
    """Util for importing objects from import path.

    :param import_path: import path of object to be imported e.g. module.submodule.Class
    :type import_path: string
    :returns: imported object
    :rtype: object
    :raises: ValueError, ImportError, AttributeError

    """
    if not (isinstance(import_path, basestring) and '.' in import_path):
        raise ValueError('There must be at least one dot in import path: "%s"', import_path)
    module_name, object_name = import_path.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, object_name)


from django.core.exceptions import ValidationError
from django.test import TestCase
from django.contrib.sites.models import Site

from brevisurl.models import ShortUrl
from brevisurl import get_connection


class TestModels(TestCase):

    def test_model_save(self):
        site = Site.objects.get_current()
        connection = get_connection('brevisurl.backends.local.BrevisUrlBackend')
        short_url = ShortUrl()
        short_url.original_url = 'http://www.codescale.net/'
        short_url.shortened_url = '{0}://{1}/12345'.format(connection.PROTOCOL, site.domain)
        short_url.backend = connection.class_path
        short_url.save()
        self.assertIsNotNone(short_url.pk)

    def test_model_save_invalid_original_url(self):
        with self.assertRaises(ValidationError):
            self.site = Site.objects.get_current()
            self.connection = get_connection('brevisurl.backends.local.BrevisUrlBackend')
            self.short_url = ShortUrl()
            self.short_url.original_url = 'www.codescale.'
            self.short_url.shortened_url = '{0}://{1}/12345'.format(self.connection.PROTOCOL, self.site.domain)
            self.short_url.backend = self.connection.class_path
            self.short_url.save()

    def test_model_save_invalid_shortened_url(self):
        with self.assertRaises(ValidationError):
            connection = get_connection('brevisurl.backends.local.BrevisUrlBackend')
            short_url = ShortUrl()
            short_url.original_url = 'http://www.codescale.net/'
            short_url.shortened_url = 'www.codescale.'
            short_url.backend = connection.class_path
            short_url.save()